import logging

from filerohr import __version__
from filerohr.pipeline import Job, pipeline_task

logger = logging.getLogger(__name__)


@pipeline_task(advertise=False)
async def version(job: Job):
    """
    Print the current filerohr version.
    """
    logger.info("Running pipeline %s with filerohr v%s", job.pipeline.id, __version__)
    job.next()



import datetime
import zoneinfo


def get_zoneinfo(tz: str | None = None):
    if tz:
        return zoneinfo.ZoneInfo(tz)
    else:
        from filerohr.config import TZ

        return zoneinfo.ZoneInfo(TZ)


def as_timezone(dt: datetime.datetime, *, tz: str | None = None):
    return dt.astimezone(get_zoneinfo(tz))


def now():
    return datetime.datetime.now(get_zoneinfo())



from filerohr.pipeline.models import AudioMetadata


def test_merge_two_audio_metadata_without_overwrite():
    meta_1 = AudioMetadata(artist="artist 1")
    meta_2 = AudioMetadata(artist="artist 2", title="title 2")
    merged = meta_1.merge(meta_2)
    assert merged is not meta_1
    assert merged is not meta_2
    assert merged.artist == "artist 1"
    assert merged.title == "title 2"


def test_merge_two_audio_metadata_with_overwrite():
    meta_1 = AudioMetadata(artist="artist 1", album="album 1")
    meta_2 = AudioMetadata(artist="artist 2", title="title 2")
    merged = meta_1.merge(meta_2, overwrite=True)
    assert merged is not meta_1
    assert merged is not meta_2
    assert merged.artist == "artist 2"
    assert merged.title == "title 2"
    assert merged.album == "album 1"



from collections.abc import Sequence
import logging
import re
from subprocess import CalledProcessError
from typing import Any, Literal

from pydantic import BaseModel, Field, field_validator, model_validator

from filerohr.pipeline import Job, Stop, pipeline_task
from filerohr.pipeline.models import AudioMetadata, LocalAudioFile, LocalAVFile, LocalFile
from filerohr.tasks import av_utils
from filerohr.tasks.av_utils import EncoderSpec
from filerohr.utils import pick

logger = logging.getLogger(__name__)


class FFMPEG(BaseModel):
    args: Sequence[str] = Field(
        description=(
            "FFMPEG command line arguments. Must contain `{input_file}` and `{output_file}` as "
            "literal placeholder strings for actual file paths."
        ),
        examples=[["-i", "{input_file}", "-vn", "{output_file}"]],
    )
    output_format: str = Field(
        pattern=re.compile(r"\.[a-z0-9]+"),
        description="Output file format. Must be a valid file extension understood by ffmpeg.",
        examples=[".mp3", ".flac"],
    )

    @field_validator("args")
    @staticmethod
    def _validate_args(args: Sequence[str]):
        if "{input_file}" not in args:
            raise ValueError("Missing input file")
        if "{output_file}" not in args:
            raise ValueError("Missing output file")
        if "-progress" in args:
            raise ValueError("Progress handling is injected by the pipeline")


@pipeline_task
async def ffmpeg(job: Job, config: FFMPEG, file: LocalFile):
    """
    Run a custom ffmpeg command with args specified in the job configuration.
    Ensure that you include `{input_file}` and `{output_file}` placeholders.
    """
    try:
        converted_path = await av_utils.ffmpeg(
            file.path,
            *config.args,
            ext=config.output_format,
            on_progress=job.on_progress,
        )
    except CalledProcessError as exc:
        raise Stop("Could not convert file with ffmpeg") from exc
    job.next(file.clone(path=converted_path))


class ExtractAudio(BaseModel):
    count: int = Field(
        default=0,
        description=(
            "Maximum number of audio streams to extract. "
            "This is useful for files that contain multiple audio streams. "
            "Extracts all audio streams if number is `0`."
        ),
    )


@pipeline_task
async def extract_audio(job: Job, config: ExtractAudio, file: LocalFile):
    """
    Extracts all audio streams in the job’s file and saves
    them as separate files.

    In case more than one stream is found, this job will
    spawn a subsequent job for each stream file.

    If you only want to extract a single stream,
    set `count: 1` in the job configuration.
    """
    counter = 0
    try:
        async for result in av_utils.get_audio_streams(
            file.path, "any", on_progress=job.on_progress
        ):
            stream_index, codec = pick(result.metadata, "stream_index", "codec", default=None)
            job.attach_logs(
                f"Extracting stream #{stream_index} with codec {codec} from {file.path}."
            )
            job.attach_logs(*result.logs())
            job.next(file.clone(LocalAudioFile, path=result.output_file_path))
            if config.count > 0:
                counter += 1
                if counter >= config.count:
                    break
    except av_utils.FFMPEGError as exc:
        job.attach_logs(*exc.logs())
        raise


class ConvertAudio(BaseModel):
    allowed_formats: list[str] | Literal["any"] = Field(
        default_factory=lambda: ["mp3", "flac", "vorbis", "opus"],
        description="List of allowed audio formats. Use 'any' to allow any audio format.",
        examples=[["vorbis", "flac", "opus"], "any"],
    )
    fallback_format: str = Field(
        default="flac",
        description=(
            "The format to fallback to if the detected format is not allowed. "
            "ffmpeg often has different encoders for the same audio format. "
            "Some of these encoders are experimental, e.g. you probably want to use"
            "`libopus` over `opus`. "
            "If you select an experimental encoder, you might have to add `['-strict', '-2']` "
            "to `fallback_format_encoder_args` to avoid errors."
        ),
    )
    fallback_format_ext: str | None = Field(
        default=None,
        description=(
            "The file extension of the fallback format. "
            "This is automatically inferred for the most common audio formats."
        ),
    )
    fallback_format_encoder_args: list[str] = Field(
        default_factory=list,
        description="Additional arguments passed to ffmpeg when encoding to the fallback format.",
    )

    @model_validator(mode="before")
    @classmethod
    def _check_fallback_format_ext(cls, data: Any):
        if not isinstance(data, dict):
            return data
        ext = data.get("fallback_format_ext")
        if ext is not None:
            return data
        fallback_format = data.get("fallback_format")
        if fallback_format in ("flac", "mp3"):
            ext = fallback_format
        elif fallback_format in ("opus", "libopus", "vorbis", "libvorbis"):
            ext = "ogg"
        else:
            raise ValueError(
                f"Cannot determine file extension for format `{fallback_format}`. "
                f"Please set `fallback_format_ext` manually."
            )
        data["fallback_format_ext"] = ext
        return data


@pipeline_task
async def convert_audio(job: Job, config: ConvertAudio, file: LocalAudioFile):
    """
    Ensures an audio file matches the allowed formats.

    Audio files that don’t match the formats will be converted
    to the fallback format.

    Note: This job must be placed after an `extract_audio` job.
    """
    encoder = EncoderSpec(
        codec=config.fallback_format,
        ext=config.fallback_format_ext,
        args=config.fallback_format_encoder_args,
    )
    try:
        async for result in av_utils.get_audio_streams(
            file.path,
            set(config.allowed_formats),
            encoder,
            on_progress=job.on_progress,
        ):
            job.attach_logs(*result.logs())
            job.next(file.clone(path=result.output_file_path))
    except av_utils.FFMPEGError as exc:
        job.attach_logs(*exc.logs())
        raise


class SanitizeAV(BaseModel):
    error_handling: Literal["ignore", "ignore_minor"] = Field(
        default="ignore",
        description=(
            "How to handle errors during file sanitization of broken files. "
            "`ignore` will ignore all but critical errors in the audio and "
            "corresponds to ffmpeg’s `-err_detect ignore_err`. "
            "The resulting file will play, but may not be pleasant to listen to. "
            "`ignore_minor` will let minor errors pass and corresponds to "
            "ffmpeg’s `-err_detect careful`."
        ),
        examples=["ignore", "ignore_minor"],
    )
    skip_invalid: bool = Field(
        default=True,
        description="Silently skip files that do not contain audio.",
    )


@pipeline_task
async def sanitize_av(job: Job, config: SanitizeAV, file: LocalFile):
    """
    Sanitizes the job file as an audio/video stream.
    This performs a simple copy operation for all streams in the
    job’s file and ensures that the resulting file is playable.

    If the file is not an audio file, it will be skipped by default.
    """
    try:
        result = await av_utils.sanitize_av_file(
            file.path, config.error_handling, on_progress=job.on_progress
        )
    except av_utils.FFMPEGError as exc:
        job.attach_logs(*exc.logs())
        if exc.process.returncode == 183 and config.skip_invalid:
            raise Stop("File does not contain audio") from exc
        raise
    else:
        job.attach_logs(*result.logs())
        job.next(file.clone(LocalAVFile, path=result.output_file_path))


@pipeline_task
async def extract_audio_metadata(job: Job, file: LocalFile):
    """
    Extracts metadata, like duration, artist, title, album
    from the job’s file.
    """
    metadata = await av_utils.get_audio_metadata(file.path, job.context.pop(AudioMetadata, None))
    duration = await av_utils.get_stream_duration(file.path, None)
    job.next(file.clone(LocalAudioFile, metadata=metadata | AudioMetadata(duration=duration)))


class NormalizeAudio(BaseModel):
    preset: Literal["music", "podcast", "streaming-video"] | None = Field(
        default=None,
        description=(
            "The audio normalization preset to use. "
            "See https://slhck.info/ffmpeg-normalize/usage/presets/ for available presets. "
            "Mutually exclusive with `args`."
        ),
    )
    args: list[str] = Field(
        default_factory=list,
        description=(
            "ffmpeg-normalize command line arguments. "
            "See https://slhck.info/ffmpeg-normalize/usage/cli-options/ for available options. "
            "Mutually exclusive with `preset`."
        ),
    )


@pipeline_task
async def normalize_audio(job: Job, config: NormalizeAudio, file: LocalAudioFile):
    """
    Normalizes the audio stream with [ffmpeg-normalize](https://slhck.info/ffmpeg-normalize/).

    If no additional options are given, the `podcast` preset will be used.

    Note: This job must be placed after an `extract_audio` job.
    """
    args = []
    if config.preset or (config.preset is None and not config.args):
        args.extend(["--preset", config.preset or "podcast"])
    if config.args:
        args.extend(config.args)
    try:
        result = await av_utils.ffmpeg_normalize(file.path, *args, on_progress=job.on_progress)
    except av_utils.FFMPEGError as exc:
        job.attach_logs(*exc.logs())
        raise
    else:
        job.attach_logs(*result.logs())
        job.next(file.clone(path=result.output_file_path))



import httpx
from pydantic import AnyHttpUrl


class HttpUrlBuilder(httpx.URL):
    def __init__(
        self, url: httpx.URL | AnyHttpUrl | str, *, use_trailing_slash: bool = False, **kwargs
    ) -> None:
        if use_trailing_slash:
            url = httpx.URL(str(url))
            if not url.path.endswith("/"):
                url = url.copy_with(path=f"{url.path}/")
        super().__init__(str(url), **kwargs)
        self._use_trailing_slash = use_trailing_slash

    def __truediv__(self, other: str | int):
        sub_paths = str(other).split("/")
        new_path = self.path
        for path in sub_paths:
            if not new_path.endswith("/"):
                new_path += "/"
            new_path += path
        if self._use_trailing_slash:
            new_path += "/"
        return self.copy_with(path=new_path)

    def copy_with(self, **kwargs) -> "HttpUrlBuilder":
        cls = type(self)
        return cls(super().copy_with(**kwargs), use_trailing_slash=self._use_trailing_slash)

    @property
    def origin(self):
        origin = f"{self.scheme}://{self.host}"
        if self.port is not None:
            origin += f":{self.port}"
        return origin



from io import StringIO
import os

import yaml

from filerohr.utils import HttpUrlBuilder, get_yaml_loader, normalize_str, omit_falsy


def test_http_url_builder_accepts_path_components():
    base_url = HttpUrlBuilder("https://example.com?test=1")
    assert str(base_url / "foo" / "bar") == "https://example.com/foo/bar?test=1"


def test_http_url_builder_appends_trailing_slashes():
    url = HttpUrlBuilder("https://example.com/test?test=1", use_trailing_slash=True)
    assert str(url) == "https://example.com/test/?test=1"


def test_http_url_builder_generates_proper_origin():
    assert HttpUrlBuilder("https://example.com/test?test=1").origin == "https://example.com"
    assert HttpUrlBuilder("http://example.com/test?test=1").origin == "http://example.com"
    assert HttpUrlBuilder("https://example.com:81/test?test=1").origin == "https://example.com:81"


def test_omit_falsy():
    def _is_not_none(value) -> bool:
        return value is not None

    assert list(omit_falsy([0, None, "", [], {}])) == []
    assert list(omit_falsy([1, "a", "b"])) == [1, "a", "b"]
    assert list(omit_falsy([0, None, ""], is_truthy=_is_not_none)) == [0, ""]


def test_normalize_str():
    # treats None as an empty string
    assert normalize_str(None) == ""
    # casts to string
    assert normalize_str(0) == "0"
    # unescapes HTML entities
    assert normalize_str("&#8211;") == "–"
    # strips whitespace
    assert normalize_str(" Hi !  ") == "Hi !"


def test_yaml_loader():
    os.environ["TEST_INT"] = "5"
    os.environ["TEST_FLOAT"] = "5.5"
    os.environ["TEST_STR"] = "test"
    os.environ["TEST_BOOL"] = "true"
    os.environ["TEST_EMPTY"] = ""
    yaml_data = StringIO(
        """
        my_int: !env TEST_INT
        my_float: !env TEST_FLOAT
        my_str: !env TEST_STR
        my_bool: !env TEST_BOOL
        my_empty: !env TEST_EMPTY
        my_missing: !env TEST_MISSING
        my_default_match: !env { name: TEST_STR, default: "hello" }
        my_default_no_match: !env { name: TEST_MISSING, default: "hello" }
        my_default_no_match_type: !env { name: TEST_MISSING, default: 10 }
        """
    )
    data = yaml.load(yaml_data, Loader=get_yaml_loader())
    expected_data = {
        "my_int": 5,
        "my_float": 5.5,
        "my_str": "test",
        "my_bool": True,
        "my_empty": "",
        "my_missing": None,
        "my_default_match": "test",
        "my_default_no_match": "hello",
        "my_default_no_match_type": 10,
    }
    assert data == expected_data



from collections.abc import Callable, Hashable, Iterable, Mapping
import functools
from html import unescape
from typing import Any

_sentinel = object()


def get_qualified_cls_name(obj) -> str:
    cls = obj if isinstance(obj, type) else type(obj)
    cls_name = cls.__qualname__
    module = cls.__module__
    if not module or module == "builtins":
        return cls_name
    if module == "types":
        return str(obj)
    return f"{module}.{cls_name}"


def once[**Params, Result](func: Callable[Params, Result]):
    sentinel = object()
    result = sentinel

    @functools.wraps(func)
    def wrapper(*args: Params.args, **kwargs: Params.kwargs) -> Result:
        nonlocal result
        if result is sentinel:
            result = func(*args, **kwargs)
        return result

    return wrapper


def omit_falsy(iterable: Iterable[Any], *, is_truthy: Callable[[Any], bool] | None = None):
    is_truthy = is_truthy or bool
    for item in iterable:
        if is_truthy(item):
            yield item


def normalize_str(value) -> str:
    if value is None:
        return ""
    return unescape(str(value)).strip()


def pick[T: Hashable](mapping: Mapping[T, Any], *keys: T, default=_sentinel):
    for key in keys:
        value = mapping.get(key, default)
        if value is _sentinel:
            raise KeyError(key)
        yield value



import logging
from pathlib import Path
import shutil

from filerohr.env import Environment

logger = logging.getLogger(__name__)

base_env = Environment()
env = Environment("FILEROHR_")

BASE_DIR = Path(__file__).parent.parent

TZ = base_env.get("TZ", default="Europe/Vienna")
FFMPEG_BIN = env.get("FFMPEG_BIN", cast=Path, default=Path("/usr/bin/ffmpeg"))
FFPROBE_BIN = env.get("FFPROBE_BIN", cast=Path, default=Path("/usr/bin/ffprobe"))
FFMPEG_NORMALIZE_BIN = env.get(
    "FFMPEG_NORMALIZE_BIN", cast=Path, default=Path(shutil.which("ffmpeg-normalize"))
)
DATA_DIR = env.get("DATA_DIR", cast=Path, default=BASE_DIR / "data")
TMP_DIR = env.get("TMP_DIR", cast=Path, default=DATA_DIR / "tmp")
TASK_MODULES = env.get_list("TASK_MODULES", default=[])
PIPELINE_CONFIG_DIR = env.get(
    "PIPELINE_CONFIG_DIR", cast=Path, default=BASE_DIR / "pipeline.conf.d"
)

DATA_DIR.mkdir(parents=True, exist_ok=True)
TMP_DIR.mkdir(parents=True, exist_ok=True)
FFMPEG_BIN.stat()
FFPROBE_BIN.stat()

if TMP_DIR.stat().st_dev != DATA_DIR.stat().st_dev:
    logger.warning("TMP_DIR on a different filesystem than DATA_DIR is discouraged.")



import os

import yaml


def _get_env_value(loader: yaml.Loader, name: str, *, default=None):
    value = os.environ.get(name, None)
    if value is None:
        return default
    if value == "":
        return value
    tag = loader.resolve(yaml.nodes.ScalarNode, value, (True, False))
    return loader.construct_object(yaml.nodes.ScalarNode(tag, value))


def _env_tag(loader: yaml.Loader, node: yaml.Node):
    if isinstance(node, yaml.nodes.ScalarNode):
        return _get_env_value(loader, loader.construct_scalar(node))

    if isinstance(node, yaml.nodes.MappingNode):
        mapping = loader.construct_mapping(node)
        try:
            _name = mapping["name"]
        except KeyError:
            raise yaml.constructor.ConstructorError(
                None,
                None,
                f'Missing "name" key in {node.id}',
                node.start_mark,
            )
        _default = mapping.get("default", None)
        return _get_env_value(loader, _name, default=_default)

    raise yaml.constructor.ConstructorError(
        None,
        None,
        f"Expected a scalar or mapping node, but found {node.id}.",
        node.start_mark,
    )


def get_yaml_loader():
    class Loader(yaml.SafeLoader):
        pass

    Loader.add_constructor("!env", _env_tag)
    return Loader



import argparse
import asyncio
import logging
import os
import sys
import textwrap

from termcolor import colored, cprint

from filerohr import config
from filerohr.pipeline import File, Job, Pipeline, get_tasks, pipeline_progress


def _show_progress(name: str, progress: float | int):
    # A rather unintelligent progress bar implementation.
    name = f"\r{name}: "
    percentage = f" {round(progress):3}% "
    bar_size = os.get_terminal_size().columns
    bar_size -= len(name)
    bar_size -= len(percentage)
    full_block_count = round(bar_size * progress / 100)
    light_block_count = bar_size - full_block_count
    bar = f"{full_block_count * '█'}{light_block_count * '░'}"

    output = name + bar + percentage
    print(output, end="", file=sys.stderr, flush=True)


def _import_file(args):
    loop = asyncio.new_event_loop()
    config = Pipeline.get_config(args.config)
    pipeline = Pipeline.create(config)
    file = File.from_path(args.file_or_url, keep=True)
    if sys.stdout.isatty() and sys.stderr.isatty():

        async def set_progress(pipeline: Pipeline, job: Job | None):
            if not job:
                return
            _show_progress(job.name, job.progress_percent or 0)

        pipeline_progress.connect(set_progress)
    processed_files = loop.run_until_complete(pipeline.start(file))
    for file in processed_files:
        print(file.model_dump_json(indent=2))
    if pipeline.status == "failed":
        sys.exit(1)
    if pipeline.status == "stopped":
        sys.exit(2)


def _show_version(args):
    from filerohr import __version__

    print(__version__)


def _serialize_type(info):
    value_type = info.get("type", None)
    if value_type == "array":
        subtype = _serialize_type(info["items"])
        if "|" in subtype:
            subtype = f"({subtype})"
        return f"{subtype}[]"
    const_value = info.get("const", None)
    if const_value:
        return repr(const_value)
    if value_type == "string" and (enum := info.get("enum", None)):
        return " | ".join(_repr(token) for token in enum)
    if value_type:
        return value_type
    any_of = info.get("anyOf", None)
    if any_of is not None:
        return " | ".join(_serialize_type(_info) for _info in any_of)
    return "unknown"


def _repr(value):
    if value is None:
        return "null"
    return repr(value)


def _list_jobs(args):
    sentinel = object()
    hpref = "#" * args.base_heading_level
    data_dir = str(config.DATA_DIR)

    first = True
    for name, creator in sorted(get_tasks()):
        if not creator.advertise:
            continue
        if not first:
            print()
        else:
            first = False
        cprint(f"{hpref}# {name}", attrs=["bold"])
        if description := creator.job_description:
            cprint(description, attrs=["dark"])
        schema = creator.config.model_json_schema()
        properties = schema["properties"]
        del properties["job"]
        if not properties:
            print()
            continue

        cprint(f"{hpref}## Configuration options:\n", attrs=[])
        for property_key, info in properties.items():
            title = colored(property_key, attrs=[])
            if value_type := _serialize_type(info):
                title += f": `{value_type}`"
            if (default := info.get("default", sentinel)) is not sentinel:
                if isinstance(default, str):
                    default = default.replace(data_dir, "$FILEROHR_DATA_DIR")
                title += f", default: `{_repr(default)}`"
            else:
                title += ", " + colored("*required*", attrs=["underline"])
            cprint(f"{hpref}### {title}")

            if description := info.get("description", ""):
                text = textwrap.fill(description)
                print()
                cprint(text, attrs=["dark"])
            if examples := info.get("examples", []):
                print()
                print("Examples:")
                for example in examples:
                    text = f"`{_repr(example)}`"
                    text = textwrap.fill(text, initial_indent="- ", subsequent_indent="  ")
                    print(text)
            print()


def _validate_pipeline_config(args):
    config = Pipeline.get_config(args.config)
    print(config.model_dump_json(indent=2))
    config.check_best_practices()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-level",
        default=logging.getLevelName(logging.WARNING),
        choices=logging.getLevelNamesMapping().keys(),
    )
    parser.add_argument("--version", action="store_const", const=_show_version)
    subparsers = parser.add_subparsers()
    process_file = subparsers.add_parser("import-file", help="Import file")
    process_file.add_argument("--config", help="Pipeline config file or name")
    process_file.add_argument("file_or_url", help="File or URL to process")
    process_file.set_defaults(func=_import_file)
    validate_pipeline_config = subparsers.add_parser(
        "validate-config", help="Loads and dumps the validated pipeline config."
    )
    validate_pipeline_config.add_argument("config", help="Pipeline config file or name")
    validate_pipeline_config.set_defaults(func=_validate_pipeline_config)
    list_jobs = subparsers.add_parser(
        "list-jobs",
        help="Show information on available jobs and their configuration options",
    )
    list_jobs.add_argument(
        "--base-heading-level",
        type=int,
        default=0,
        help="Markdown heading indent level",
    )
    list_jobs.set_defaults(func=_list_jobs)
    return parser


def main(args=None):
    parser = get_parser()
    parsed_args = parser.parse_args(args)

    logging.basicConfig(
        level=logging.getLevelNamesMapping()[parsed_args.log_level],
        handlers=[logging.StreamHandler()],
    )
    func = getattr(parsed_args, "func", None)
    if func is None:
        parser.print_help()
        sys.exit(1)
    else:
        func(parsed_args)


if __name__ == "__main__":
    main()



import datetime
from pathlib import Path, PurePath
from typing import Any
import urllib.parse
import uuid

from pydantic import BaseModel, ConfigDict, Field

from filerohr.utils import now


class Metadata(BaseModel):
    model_config = ConfigDict(extra="allow")

    def merge[T](self: type[T], other: BaseModel | dict[str, Any], overwrite: bool = False) -> T:
        all_fields = set(dict(self).keys()) | set(dict(other).keys())
        self_set_fields = self.model_fields_set
        try:
            other_set_fields = other.model_fields_set
        except AttributeError:
            other_set_fields = set(other.keys())
        iter_fields = all_fields if overwrite else all_fields - self_set_fields
        iter_fields &= other_set_fields
        try:
            update = other.model_dump(include=iter_fields)
        except AttributeError:
            update = {key: other[key] for key in iter_fields}
        return self.model_copy(update=update)

    def __or__[T](self: type[T], other: BaseModel | dict[str, Any]) -> T:
        return self.merge(other)


class File(BaseModel):
    name: str
    path: Path | str
    id: uuid.uuid4 = Field(default_factory=uuid.uuid4)
    hash: str | None = None
    mime_type: str | None = None
    created_at: datetime.datetime = Field(default_factory=now)
    metadata: Metadata = Field(default_factory=Metadata)
    keep: bool = False
    log: str = ""

    def clone(self, cls: type["File"] | None = None, **kwargs):
        if cls is None:
            cls = type(self)
        new_attrs = {"keep": False} if "path" in kwargs else {}
        new_attrs.update(kwargs)
        data = self.model_copy(deep=True).model_dump()
        data.update(new_attrs)
        del data["id"]
        if "path" in kwargs and "hash" not in kwargs:
            del data["hash"]
        return cls(**data)

    @classmethod
    def from_path(cls, path: Path | str, **kwargs):
        path_str = str(path)
        url = urllib.parse.urlparse(path_str)
        if url.scheme:
            return cls.from_url(path_str, **kwargs)
        return cls.from_filesystem(Path(path), **kwargs)

    @staticmethod
    def from_filesystem(path: Path, **kwargs):
        kwargs.setdefault("name", path.stem)
        return LocalFile(path=path, **kwargs)

    @staticmethod
    def from_url(url: str, **kwargs):
        kwargs.setdefault("name", PurePath(urllib.parse.urlparse(url).path).name)
        return RemoteFile(path=url, **kwargs)


class RemoteFile(File):
    path: str


class LocalFile(File):
    path: Path


class AVMetadata(Metadata):
    duration: datetime.timedelta | None = Field(default=None)


class LocalAVFile(LocalFile):
    metadata: AVMetadata = Field(default_factory=AVMetadata)


class AudioMetadata(AVMetadata):
    album: str = ""
    artist: str = ""
    isrc: str = ""
    organization: str = ""
    title: str = ""
    date: str = ""
    genre: str = ""


class LocalAudioFile(LocalAVFile):
    metadata: AudioMetadata = Field(default_factory=AudioMetadata)



import asyncio
from collections import defaultdict
from collections.abc import AsyncIterable, Mapping
import contextlib
import datetime
import json
import logging
from pathlib import Path
import re
from typing import Any, Literal

import aiofiles
from pydantic import BaseModel

from filerohr import config
from filerohr.pipeline.models import AudioMetadata
from filerohr.pipeline.types import OnProgress
from filerohr.utils import create_fifo, create_tmp_file, normalize_str

logger = logging.getLogger(__name__)


class FFMPEGMixin:
    process: asyncio.subprocess.Process
    cmd: list[str]
    stdout: str
    stderr: str
    metadata: Mapping[str, Any]

    def __init__(
        self,
        cmd: list[str],
        process: asyncio.subprocess.Process,
        *,
        metadata: Mapping[str, Any] | None = None,
        stdout: str | None = None,
        stderr: str | None = None,
    ):
        self.cmd = cmd
        self.process = process
        self.stdout = stdout or ""
        self.stderr = stderr or ""
        self.metadata = metadata or {}
        super().__init__()

    @property
    def cmd_str(self):
        return " ".join(self.cmd)

    def logs(self):
        yield self.cmd_str
        yield self.stdout
        yield self.stderr


class FFMPEGResult(FFMPEGMixin):
    output_file_path: Path

    def __init__(self, output_file_path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_file_path = output_file_path


class FFMPEGError(FFMPEGMixin, RuntimeError):
    pass


class EncoderSpec(BaseModel):
    codec: str
    ext: str
    args: list[str]


async def ffprobe(*args: str) -> str:
    cmd = [str(config.FFPROBE_BIN), "-output_format", "json", *args]
    logger.debug("Starting ffprobe with args: %s", " ".join(cmd))
    process = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await process.communicate()
    stdout = stdout.decode(errors="replace")
    stderr = stderr.decode(errors="replace")
    if process.returncode != 0:
        raise FFMPEGError(cmd, process, stdout=stdout, stderr=stderr)
    return stdout


async def ffmpeg(
    path: Path,
    *args,
    ext: str | None = None,
    on_progress: OnProgress | None = None,
    stream_index: int | None = None,
    metadata: Mapping[str, Any] | None = None,
):
    injected_args = ["-y", "-bitexact"]

    if ext is None:
        ext = path.suffix
    elif not ext.startswith("."):
        ext = f".{ext}"

    async with contextlib.AsyncExitStack() as stack:
        tmp_file_path = await stack.enter_async_context(
            create_tmp_file(suffix=ext, delete=False, dir=config.TMP_DIR)
        )

        fifo_path: Path | None = None
        duration: datetime.timedelta | None = None
        if on_progress:
            logger.debug("Creating FFMPEG progress pipe")
            duration = await get_stream_duration(path, stream_index)
        # we can only calculate the progress if we have a duration
        if duration:
            fifo_path = await stack.enter_async_context(
                create_fifo(tmp_file_path.with_suffix(".fifo"))
            )
            injected_args.extend(["-progress", str(fifo_path), "-stats_period", "0.1"])

        args = [arg.format(input_file=path, output_file=str(tmp_file_path)) for arg in args]
        cmd = [str(config.FFMPEG_BIN), *injected_args, *args]
        logger.debug("Starting ffmpeg with args: %s", " ".join(cmd))
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )

        if fifo_path and duration:
            duration_micro_seconds = int(duration.total_seconds() * 1000 * 1000)
            logger.debug("Opening FFMPEG progress pipe")
            async with aiofiles.open(fifo_path, "r") as fifo:
                async for line in fifo:
                    key, value = line.strip().split("=")
                    if key == "progress" and value == "end":
                        logger.debug("FFMPEG finished work")
                        break
                    if key == "out_time_us":
                        # out_time_us is microseconds
                        try:
                            on_progress(float(value) / duration_micro_seconds * 100)
                        except ValueError:
                            pass
            logger.debug("FFMPEG progress pipe closed")

        stdout, _ = await process.communicate()
    logs = stdout.decode(errors="replace")
    if process.returncode != 0:
        raise FFMPEGError(cmd, process, stdout=logs, metadata=metadata)
    return FFMPEGResult(tmp_file_path, cmd, process, stdout=logs, metadata=metadata)


async def _handle_ffmpeg_normalize_progress(
    args: tuple[str, ...],
    process: asyncio.subprocess.Process,
    on_progress: OnProgress,
):
    min_passes = 1 if "--dynamic" in args else 2
    pass_progress: dict[str, float] = defaultdict(float)
    progress_pattern = re.compile(r"^(.+): (\d+\.\d+)%")
    stdout = ""
    async for line in process.stdout:
        text = line.decode(errors="replace").strip()
        stdout += text + "\n"
        if match := re.match(progress_pattern, text):
            name, percentage = match.groups()
            # sometimes the name includes the progress bar
            if "█" in name:
                continue
            # a new pass indicates that the previous pass has finished
            if name not in pass_progress:
                for _name in pass_progress.keys():
                    pass_progress[_name] = 100
            pass_progress[name] = float(percentage)
            on_progress(sum(pass_progress.values()) / max(min_passes, len(pass_progress)))
    on_progress(100)
    return stdout


async def ffmpeg_normalize(
    path: Path,
    *args: str,
    on_progress: OnProgress | None = None,
    metadata: Mapping[str, Any] | None = None,
):
    async with contextlib.AsyncExitStack() as stack:
        tmp_file_path = await stack.enter_async_context(
            create_tmp_file(suffix=".wav", delete=False, dir=config.TMP_DIR)
        )
        # fmt: off
        cmd = [
            str(config.FFMPEG_NORMALIZE_BIN),
            str(path),
            "--force",
            *args,
            "--output", str(tmp_file_path)
        ]
        # fmt: on
        logger.debug("Starting ffmpeg-normalize with args: %s", " ".join(cmd))
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            env={
                "FFMPEG_PATH": str(config.FFMPEG_BIN),
                "TMP": str(config.TMP_DIR / "ffmpeg-normalize"),
            },
        )
        stdout = await _handle_ffmpeg_normalize_progress(args, process, on_progress)
        await process.wait()
    if process.returncode != 0:
        raise FFMPEGError(cmd, process, stdout=stdout, metadata=metadata)
    return FFMPEGResult(tmp_file_path, cmd, process, stdout=stdout, metadata=metadata)


async def sanitize_av_file(
    path: Path,
    error_handling: Literal["ignore", "ignore_minor"],
    on_progress: OnProgress | None = None,
    **kwargs,
) -> FFMPEGResult:
    if error_handling == "ignore":
        err_detect = "ignore_err"
    elif error_handling == "ignore_minor":
        err_detect = "careful"
    else:
        raise ValueError(f"Invalid error handling: {error_handling}")

    # fmt: off
    return await ffmpeg(
        path,
        "-err_detect", err_detect,
        "-i", "{input_file}",
        # we copy all streams, this is just about sanitization
        "-c", "copy",
        "-map", "0",
        "{output_file}",
        on_progress=on_progress,
        **kwargs,
    )
    # fmt: on


async def get_audio_stream_codecs(path: Path) -> AsyncIterable[tuple[int, str]]:
    # fmt: off
    output = await ffprobe(
        # only select audio streams
        "-select_streams", "a",
        # only return the stream index and codec name
        "-show_entries", "stream=index,codec_name",
        str(path),
    )
    # fmt: on
    result = json.loads(output)
    streams = result.get("streams", [])
    if not streams:
        return
    for stream in streams:
        codec = stream.get("codec_name", None)
        if codec is None:
            continue
        yield stream["index"], codec


async def _extract_audio_stream(
    path: Path, stream_index: int, encoder: EncoderSpec | None = None, **kwargs
):
    codec = encoder.codec if encoder else "copy"
    ext = encoder.ext if encoder else path.suffix
    args = encoder.args if encoder else []
    if codec == "copy":
        logger.debug("Keeping codec for stream %d in `%s`.", stream_index, path)
    else:
        logger.debug("Re-encoding stream %d in `%s` as `%s`.", stream_index, path, codec)
    # fmt: off
    return await ffmpeg(
        path,
        "-i", "{input_file}",
        # drop video streams
        "-vn",
        # keep metadata
        "-map_metadata", "0",
        # select the specified audio stream
        "-map", f"0:a:{stream_index}",
        # convert to the specified codec
        "-c:a", codec,
        *args,
        "{output_file}",
        ext=ext,
        stream_index=stream_index,
        **kwargs,
    )
    # fmt: on


async def get_audio_streams(
    path: Path,
    allowed_formats: set[str] | Literal["any"],
    encoder: EncoderSpec | None = None,
    on_progress: OnProgress | None = None,
):
    stream_progress: list[float] = []
    streams: list[tuple[int, str]] = []
    async for stream in get_audio_stream_codecs(path):
        streams.append(stream)
        stream_progress.append(0)

    for index, [stream_index, codec] in enumerate(streams):

        def _on_stream_progress(progress):
            stream_progress[index] = progress
            on_progress(sum(stream_progress) / len(stream_progress))

        metadata = {"index": index, "stream_index": stream_index, "codec": codec}
        _on_progress = _on_stream_progress if on_progress else None
        if allowed_formats == "any" or codec in allowed_formats:
            encoder = None
        elif encoder is None:
            raise ValueError(
                f"Audio stream uses codec {codec}, which is not allowed, "
                f"but no encoder was provided."
            )
        yield await _extract_audio_stream(
            path,
            stream_index,
            encoder,
            on_progress=_on_progress,
            metadata=metadata,
        )


async def get_stream_durations(path: Path):
    output = await ffprobe("-show_entries", "stream=index,duration", str(path))
    ffprobe_data = json.loads(output)
    result = {}
    for stream in ffprobe_data.get("streams", []):
        try:
            duration = float(stream.get("duration", None))
        except (KeyError, TypeError):
            continue
        result[stream["index"]] = datetime.timedelta(seconds=duration)
    return result


async def get_stream_duration(path: Path, stream_index: int | None):
    durations = await get_stream_durations(path)
    if stream_index is None:
        try:
            return list(durations.values())[0]
        except IndexError:
            return None
    try:
        return durations[stream_index]
    except KeyError:
        return None


async def get_audio_metadata(
    path: Path,
    existing_metadata: AudioMetadata | None = None,
) -> AudioMetadata:
    # fmt: off
    output = await ffprobe(
        # only select audio streams
        "-select_streams", "a",
        # extract stream and format tags
        "-show_entries", "stream_tags:format_tags",
        str(path),
    )
    # fmt: on
    result = json.loads(output)
    format_tags = result.get("format", {}).get("tags", {})
    metadata = AudioMetadata(
        title=normalize_str(format_tags.get("title", "")),
        artist=normalize_str(format_tags.get("artist", "")),
        album=normalize_str(format_tags.get("album", "")),
        date=normalize_str(format_tags.get("date", "")),
        genre=normalize_str(format_tags.get("genre", "")),
    )
    if existing_metadata:
        metadata = existing_metadata.merge(metadata)
    streams = result.get("streams", [])
    if len(streams) > 1:
        logger.warning(
            "More than one audio stream found in file. Extract streams before extracting metadata."
        )
        return metadata
    stream = streams[0]
    # if there are no stream tags to process, we’re done
    if "tags" not in stream or not stream["tags"]:
        return metadata
    stream_tags = {key.lower(): value for key, value in stream["tags"].items()}
    # check if we have any unpopulated in the metadata that
    # we should populate from the stream tags
    unpopulated_keys = set(key for key, value in metadata.model_dump().items() if value == "")
    if not unpopulated_keys:
        return metadata
    for key, value in stream_tags.items():
        normalized_value = normalize_str(value)
        if normalized_value and key in unpopulated_keys:
            setattr(metadata, key, normalized_value)
            unpopulated_keys.remove(key)
    return metadata



import logging
from pathlib import Path
from typing import Literal

from httpx import HTTPError
from pydantic import BaseModel, Field

from filerohr import config as filerohr_config
from filerohr.pipeline import Job, SkipJob, Stop, pipeline_task
from filerohr.pipeline.models import LocalFile, RemoteFile
from filerohr.tasks import remote_utils
from filerohr.tasks.remote_utils import DisallowedSchemeError

logger = logging.getLogger(__name__)


class HTTPDownload(BaseModel):
    storage_dir: Path = Field(
        default=filerohr_config.TMP_DIR,
        description="Base directory to store downloaded files in.",
    )
    max_download_size_mib: float = Field(
        default=1024,
        description=(
            "Maximum size allowed for downloaded files in MiB (megabytes). Use `0` for no limit."
        ),
    )
    allow_streams: bool = Field(
        default=False,
        description=(
            "Whether to download files from sources that are streaming responses "
            "and do not have a fixed file size."
        ),
    )
    allowed_protocols: list[Literal["http", "https"]] = Field(
        default=["http", "https"],
        description="List of allowed protocols to download files from.",
    )
    timeout_seconds: int = Field(
        default=600,
        description="Timeout in seconds for downloading.",
    )
    follow_redirects: bool = Field(
        default=True,
        description="Follow redirects when downloading.",
    )
    match_content_type: list[str | Literal["pass_unset"]] | None = Field(
        default=None,
        description=(
            "List of mime types to check. Supports glob patterns like `audio/*`. "
            "Add `pass_unset` to the list if you want this check to pass "
            "in case no Content-Type was given in the server’s response headers. "
            "Setting `match_content_type` to `null` will allow all content types."
        ),
    )
    download_chunk_size_mib: float = Field(
        default=1,
        description="Size of chunks to download in MiB (megabytes).",
    )


@pipeline_task
async def download_file(job: Job, config: HTTPDownload):
    """
    Downloads files from remote sources.

    When called with a local file path, the job will simply be skipped.
    """
    if not isinstance(job.file, RemoteFile):
        raise SkipJob()

    file = job.file
    url: str = file.path
    try:
        path, name, mime_type = await remote_utils.download_file(
            url,
            follow_redirects=config.follow_redirects,
            timeout_seconds=config.timeout_seconds,
            match_content_type=config.match_content_type,
            max_download_size_mib=config.max_download_size_mib,
            allow_streams=config.allow_streams,
            storage_dir=config.storage_dir,
            download_chunk_size_mib=config.download_chunk_size_mib,
            on_progress=job.on_progress,
        )
    except DisallowedSchemeError:
        logger.info("Skipping download of %s: Disallowed scheme.", url)
        raise SkipJob()
    except (TimeoutError, ValueError, HTTPError) as exc:
        logger.info("Stopping download of %s: %s", url, str(exc))
        raise Stop() from exc
    job.next(job.file.clone(LocalFile, path=path, name=file.name or name, mime_type=mime_type))



import asyncio
from collections.abc import Awaitable, Callable
import dataclasses
import datetime
import functools
import importlib
import inspect
import itertools
import logging
import os.path
from pathlib import Path
import time
from typing import Annotated, ClassVar, List, Literal, Optional, Union, cast
import uuid

import aiofiles.os
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PrivateAttr,
    create_model,
)
import yaml

from filerohr.pipeline.events import file_emit, pipeline_progress
from filerohr.pipeline.models import File
from filerohr.pipeline.types import OperationStatus
from filerohr.utils import cast_bool, get_qualified_cls_name, get_yaml_loader, now, once

logger = logging.getLogger(__name__)

type TaskExecutor = Callable[[Job], Awaitable[None]]


class Stop(Exception):
    """
    Stop the pipeline
    """

    def __str__(self):
        return super().__str__() or str(self.__cause__)


class PipelineConfigurationError(ValueError):
    pass


class JobConfig(BaseModel):
    job: str
    skip: Annotated[bool, BeforeValidator(cast_bool)] = False

    model_config = ConfigDict(frozen=True, extra="forbid")

    def keeps_file(self):
        return False


class PipelineConfig(BaseModel):
    jobs: list[JobConfig]
    name: str | None = None
    use_as_default: bool = False

    model_config = ConfigDict(frozen=True, extra="forbid")

    def check_best_practices(self):
        if not any(job.keeps_file() for job in self.jobs):
            logger.warning(
                "The pipeline config does not contain any job that will keep their output file. "
                "No files will be stored after the pipeline finishes."
            )


class Runnable(BaseModel):
    started_at: datetime.datetime | None = Field(default=None, init=False)
    finished_at: datetime.datetime | None = Field(default=None, init=False)
    duration: datetime.timedelta | None = Field(default=None, init=False)


class Pipeline(Runnable):
    name: str | None = None
    id: uuid.uuid4 = Field(default_factory=uuid.uuid4)
    tasks: list[tuple[str, Callable[["Job"], Awaitable[None]]]]
    jobs: list["Job"] = Field(default_factory=list)
    status: OperationStatus = "created"
    _emitted_files: list[File] | list = PrivateAttr(default_factory=list)
    _queue: asyncio.Queue | None = PrivateAttr(default=None)

    _global_configs: ClassVar[dict[str, PipelineConfig]] = {}

    @classmethod
    def get_config(cls, name_or_path: str | Path | None):
        _patch_pipeline_config_type()
        if name_or_path and os.path.isfile(name_or_path):
            return load_pipeline_config(Path(name_or_path))
        load_global_pipeline_configs()
        if name_or_path is None:
            default_config = None
            for config in cls._global_configs.values():
                if config.use_as_default:
                    default_config = config
                    break
            if default_config is None:
                raise ValueError("No default pipeline configuration found.")
            return default_config
        try:
            return cls._global_configs[name_or_path]
        except KeyError:
            raise ValueError(f"No pipeline named `{name_or_path}`")

    @classmethod
    def create(cls, pipeline_config: PipelineConfig | str | None = None, **kwargs):
        if isinstance(pipeline_config, str) or pipeline_config is None:
            pipeline_config = cls.get_config(pipeline_config)
        if pipeline_config is None:
            raise ValueError("Pipeline configuration is required")
        tasks = []
        for job_config in pipeline_config.jobs:
            task_name = job_config.job
            try:
                task_creator = task_creator_map[task_name]
            except KeyError:
                raise ValueError(f"Unknown task '{task_name}'")
            task_executor = task_creator(job_config)
            tasks.append((task_name, task_executor))

        return cls(name=pipeline_config.name, tasks=tasks, **kwargs)

    async def start(self, file: File, context: dict | None = None):
        if self.status != "created":
            raise RuntimeError(
                "Pipeline was already started. Please create a new pipeline instance."
            )
        self.started_at = now()
        time_started = time.monotonic()
        self._queue = asyncio.Queue()
        self.status = "processing"
        _on_progress(self)
        self._next(None, file, context or {})
        try:
            while True:
                logger.debug("[Pipeline=%s] Waiting for next job", self.id)
                job = await self._queue.get()
                await job
                self._queue.task_done()
                if self._queue.empty():
                    break
            await self._cleanup()
        except Stop:
            self.status = "stopped"
        except Exception:
            self.status = "failed"
            raise
        else:
            self.status = "completed"
        finally:
            self.finished_at = now()
            self.duration = datetime.timedelta(seconds=time.monotonic() - time_started)
            _on_progress(self)
        return self._emitted_files

    async def _cleanup(self):
        paths = set()
        keep_paths = set()
        for job in self.jobs:
            paths.add(job.file.path)
            if job.file.keep:
                keep_paths.add(job.file.path)
        delete_paths = paths - keep_paths
        for path in delete_paths:
            try:
                await aiofiles.os.unlink(path)
            except FileNotFoundError:
                pass

    def _next(self, job: Optional["Job"], file: File, context: dict):
        if not self.tasks:
            return

        new_task_index = job.task_index + 1 if job else 0
        try:
            task_name, task = self.tasks[new_task_index]
        except IndexError:
            # that was the last task
            return

        job = Job(
            name=task_name,
            pipeline=self,
            file=file,
            task=task,
            task_index=new_task_index,
            context=context,
            predecessor=job,
        )

        def emit_file(file: File):
            self._emitted_files.append(file)
            asyncio.create_task(file_emit.send_async(self, job=job, file=file))

        def next_job(file: File, context: dict):
            self._next(job, file, context)

        self.jobs.append(job)
        job_runner = job.run(emit_file=emit_file, next_job=next_job)
        self._queue.put_nowait(asyncio.create_task(job_runner, name=task_name))


class SkipJob(Exception):
    pass


class Job(Runnable):
    id: uuid.uuid4 = Field(default_factory=uuid.uuid4)
    name: str
    pipeline: Pipeline
    file: File
    task: Callable[["Job"], Awaitable[None]]
    task_index: int = 0
    predecessor: Optional["Job"] = None
    status: OperationStatus | Literal["skipped"] = "created"
    progress_percent: float | None = None
    logs: list[str] = Field(default_factory=list)
    context: dict = Field(default_factory=dict)
    _emit_file: Callable[[File], None] = PrivateAttr()
    _next_job: Callable[[File, dict], None] = PrivateAttr()

    def attach_logs(self, *logs: str | None):
        for log in logs:
            log = log.strip() if log else ""
            if not log:
                continue
            self.logs.append(log)

    def on_progress(self, progress: float):
        if not isinstance(progress, (float, int)):
            logger.error("Job %s produces invalid progress: %s", self.name, str(progress))
            return
        if self.status != "processing":
            return
        progress = min(max(progress, 0), 100)
        if progress == self.progress_percent:
            return
        logger.debug(
            "[Pipeline=%s] Job %s progress: %.2f%%", self.pipeline.id, self.name, progress
        )
        self.progress_percent = float(progress)
        _on_progress(self)

    def merge_logs(self):
        """
        Recursively merges the logs of this job and all its predecessors.
        """
        job = self
        logs = []
        while True:
            job_logs = []
            for log in job.logs:
                job_logs.append(log)
            logs.append("\n".join(job_logs))
            job = job.predecessor
            if job is None:
                break
        # job logs are in order, logs across jobs are in reverse order
        return "\n".join(reversed(logs))

    def emit_file(self, file: File):
        file = file.clone(log=self.merge_logs(), keep=True)
        self._emit_file(file)
        return file

    def next(
        self,
        file: File | None = None,
        context: dict | None = None,
        *,
        emit: bool = False,
    ):
        if file is None:
            file = self.file
        if emit:
            file = self.emit_file(file)
        new_context = dict(self.context)
        if context:
            new_context.update(context)
        self._next_job(file, new_context)

    async def run(
        self,
        *,
        emit_file: Callable[[File], None],
        next_job: Callable[[File, dict], None],
    ):
        self._emit_file = emit_file
        self._next_job = next_job

        self.started_at = now()
        time_started = time.monotonic()
        logger.debug(
            "[Pipeline=%s] Starting job %s for file %s",
            self.pipeline.id,
            self.name,
            self.file.path,
        )
        self.status = "processing"
        _on_progress(self)
        try:
            await self.task(self)
        except Stop:
            logger.info(f"Stopping processing of file {self.file.path}.")
            self.status = "stopped"
            raise
        except SkipJob:
            self.next()
            self.status = "skipped"
        except Exception:
            if self.logs:
                for log in self.logs:
                    logger.error(log)
            self.status = "failed"
            raise
        else:
            self.status = "completed"
        finally:
            time_taken = time.monotonic() - time_started
            self.progress_percent = 100
            self.finished_at = now()
            self.duration = datetime.timedelta(seconds=time_taken)
            msg = (
                f"{self.status.capitalize()} job {self.name} for file {self.file.path}. "
                f"Took {time_taken:.2f} seconds."
            )
            logger.info("[Pipeline=%s] %s", self.pipeline.id, msg)
            self.attach_logs(msg)
            _on_progress(self)


def _on_progress(job_or_pipeline: Pipeline | Job):
    if isinstance(job_or_pipeline, Pipeline):
        pipeline = job_or_pipeline
        job = None
    else:
        pipeline = job_or_pipeline.pipeline
        job = job_or_pipeline
    asyncio.create_task(pipeline_progress.send_async(pipeline, job=job))


@dataclasses.dataclass
class TaskCreator[Config: BaseModel]:
    def __init__(
        self,
        config: type[Config],
        func: Callable[[Job, Config], Awaitable[None]],
        advertise: bool,
    ):
        self._func = func
        self._sig = inspect.signature(func)
        self.config = config
        self.job_description = func.__doc__ or ""
        self.advertise = advertise

    def __str__(self):
        return self._func.__name__

    def _get_required_file_cls(self):
        try:
            file_param = self._sig.parameters["file"]
        except KeyError:
            return None
        return file_param.annotation

    def __call__(self, config: Config):
        @functools.wraps(self._func)
        async def run_task(job: Job):
            kwargs = {}
            if "job" in self._sig.parameters:
                kwargs["job"] = job
            if "config" in self._sig.parameters:
                kwargs["config"] = config
            required_file_cls = self._get_required_file_cls()
            if required_file_cls is not None:
                # isinstance will only work for basic type expressions
                # For more complex cases we need something like typeguard’s check_type function
                # see: https://pypi.org/project/typeguard/
                if not isinstance(job.file, required_file_cls):
                    req_name = get_qualified_cls_name(required_file_cls)
                    logger.error(
                        "Job `%s` produced file of type `%s`, but `%s` expected `%s`.",
                        job.predecessor.name,
                        get_qualified_cls_name(job.file),
                        job.name,
                        req_name,
                    )
                    raise Stop(f"Job expected file of type {req_name}.")
                kwargs["file"] = job.file
            await self._func(**kwargs)

        return run_task


task_creator_map: dict[str, TaskCreator] = {}


def pipeline_task[Config: BaseModel](
    func: Callable[[Job, Config], Awaitable[None]] | None = None,
    *,
    name: str | None = None,
    advertise: bool = True,
):
    def decorator(func: Callable[[Job, Config], Awaitable[None]]):
        _name = name or func.__name__
        if _name in task_creator_map:
            raise ValueError(f"Detected duplicate task definition for '{_name}'.")

        bases = [JobConfig]
        try:
            config_model = inspect.signature(func).parameters["config"].annotation
        except KeyError:
            config_model = None
        if config_model is not None:
            bases.insert(0, config_model)

        class _JobConfig(*bases):
            job: Literal[_name]

        # generate pretty class names
        if config_model is None:
            job_config = type(f"{_name}_JobConfig", (_JobConfig,), {})
        else:
            job_config = type(f"{config_model.__name__}JobConfig", (_JobConfig,), {})

        task_creator_map[_name] = TaskCreator(cast(type[_JobConfig], job_config), func, advertise)
        return func

    if func is not None:
        return decorator(func)
    else:
        return decorator


@once
def _load_tasks():
    """
    Loads all local and external tasks.
    """
    from filerohr import (
        config,
        tasks,  # noqa: F401
    )

    for module_string in config.TASK_MODULES:
        importlib.import_module(module_string)


@once
def _patch_pipeline_config_type():
    """
    The pipeline config type uses a stub for the `jobs` field
    because we don’t know what type of jobs are available
    before the task runners have loaded.

    We therefore need to patch the type of the `jobs` field
    once we have loaded all tasks.
    """
    _load_tasks()
    job_configs = tuple(creator.config for creator in task_creator_map.values())
    final_pipeline_config = create_model(
        "FinalPipelineConfig",
        __base__=PipelineConfig,
        jobs=Annotated[
            List[
                Annotated[
                    Union[job_configs],
                    Field(discriminator="job"),
                ]
            ],
            AfterValidator(lambda jobs: [job for job in jobs if not job.skip]),
        ],
    )

    globals().update({PipelineConfig.__name__: final_pipeline_config})


def load_pipeline_config(path: Path):
    """
    Loads a pipeline configuration from a YAML file.
    """
    _patch_pipeline_config_type()
    with path.open() as f:
        try:
            raw_config = yaml.load(f, Loader=get_yaml_loader())
        except yaml.YAMLError as exc:
            msg = f"Pipeline configuration in '{path}' is not valid YAML."
            raise PipelineConfigurationError(msg) from exc
    return PipelineConfig(**raw_config)


@once
def load_global_pipeline_configs():
    from filerohr import config

    _patch_pipeline_config_type()

    config_dir = config.PIPELINE_CONFIG_DIR
    configs: dict[str, PipelineConfig] = {}
    default_config_name: str | None = None
    for path in itertools.chain(config_dir.rglob("*.yaml"), config_dir.rglob("*.yml")):
        config = load_pipeline_config(path)
        name = config.name
        if not name:
            logger.warning("Skipping pipeline configuration in `%s`. No name defined.", path)
        if config.name in configs:
            raise PipelineConfigurationError(f"Duplicate pipeline configuration name: `{name}`.")
        if config.use_as_default:
            if default_config_name is not None:
                raise PipelineConfigurationError(
                    f"Config `{name}` is configured as default but so `{default_config_name}`."
                )
            default_config_name = name
        configs[config.name] = config
    Pipeline._global_configs = configs


def get_tasks():
    _load_tasks()
    items = task_creator_map.items()
    return sorted(items, key=lambda item: item[0])



import os
import typing


class _Undefined:
    pass


def _noop(value: str) -> str:
    return value


class Environment:
    # TODO: remove all the typing.overload calls
    #       once https://github.com/python/mypy/issues/3737 has been fixed
    def __init__(self, prefix: str = ""):
        self.prefix = prefix

    def _gen_name(self, env_var_name: str):
        return f"{self.prefix}{env_var_name}"

    @typing.overload
    def get(
        self,
        env_var_name: str,
        *,
        cast: typing.Callable[[str], str] = _noop,
        default: str | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> str: ...

    @typing.overload
    def get[T](
        self,
        env_var_name: str,
        *,
        cast: typing.Callable[[str], T],
        default: T | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> T: ...

    def get[T](
        self,
        env_var_name: str,
        *,
        cast=_noop,
        default: T | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> T:
        name = self._gen_name(env_var_name)
        try:
            value = os.environ[name]
        except KeyError:
            if default is not _Undefined:
                return typing.cast(T, default)
            raise
        else:
            if pop:
                del os.environ[name]
            return cast(value)

    @typing.overload
    def get_list(
        self,
        env_var_name: str,
        *,
        separator: str = ",",
        cast: typing.Callable[[str], str] = _noop,
        default: list[str] | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> list[str]: ...

    @typing.overload
    def get_list[T](
        self,
        env_var_name: str,
        *,
        separator: str = ",",
        cast: typing.Callable[[str], T],
        default: list[T] | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> list[T]: ...

    def get_list[T](
        self,
        env_var_name: str,
        *,
        separator: str = ",",
        cast=_noop,
        default: list[T] | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> list[T]:
        try:
            value = self.get(env_var_name, pop=pop)
        except KeyError:
            if default is not _Undefined:
                return typing.cast(list[T], default)
            raise
        split_env = [v.strip() for v in value.split(separator)]
        return [cast(v) for v in split_env if v]

    def is_true(
        self,
        env_var_name: str,
        *,
        default: bool | type[_Undefined] = _Undefined,
        pop: bool = False,
    ) -> bool:
        try:
            value = self.get(env_var_name, pop=pop).lower()
        except KeyError:
            if default is not _Undefined:
                return typing.cast(bool, default)
            raise
        else:
            return value in {"1", "yes", "on", "true"}

    def __contains__(self, item):
        name = self._gen_name(item)
        return name in os.environ



import hashlib
import logging
from pathlib import Path
from typing import Literal

import aiofiles.os
from pydantic import BaseModel, Field, model_validator

from filerohr import (
    config as filerohr_config,
    utils,
)
from filerohr.pipeline import Job, pipeline_task
from filerohr.pipeline.models import LocalFile

logger = logging.getLogger(__name__)
_hashes = sorted(hashlib.algorithms_guaranteed)


@pipeline_task
async def extract_mime_type(job: Job, file: LocalFile):
    """
    Extracts the mime type of the job’s file.
    """
    mime_type = await utils.get_mime_type(file.path)
    job.next(file.clone(mime_type=mime_type))


class HashFile(BaseModel):
    alg: Literal[*_hashes] = Field(
        default="sha256",
        description="Hash algorithm to use.",
    )


@pipeline_task
async def hash_file(job: Job, config: HashFile, file: LocalFile):
    """
    Hash the file content.
    """
    file_hash = await utils.hash_file(file.path, alg=config.alg)
    job.next(file.clone(hash=file_hash))


class StoreFile(BaseModel):
    storage_dir: Path = Field(
        default=filerohr_config.DATA_DIR,
        description="Base directory to store files in.",
    )
    symlink: bool = Field(
        default=False,
        description=(
            "If set to true, a symlink is created to the file from the previous job. "
            "In that case, the previous job must create the file in a persistent storage location."
        ),
    )
    keep: bool = Field(
        default=False,
        description="If set to true, the file is kept after the pipeline finishes.",
    )
    emit: bool = Field(
        default=False,
        description="If set to true, the file is emitted as a pipeline result. Implies keep.",
    )

    def keeps_file(self):
        return self.keep or self.emit

    @model_validator(mode="after")
    @staticmethod
    def _validate_persistent_symlinks(instance: "StoreFile"):
        if instance.symlink and not (instance.keep or instance.emit):
            raise ValueError(
                "Symlinks can only be created if the result is either emitted or kept. "
                "Please set emit or keep to true."
            )
        return instance


async def store_file(job: Job, config: StoreFile, new_path: Path, **kwargs):
    stored_file = job.file.clone(path=new_path, keep=config.keep, **kwargs)
    if config.symlink:
        if not job.file.keep:
            logger.error(
                (
                    "Job %s did not create a persistent file. "
                    "Please configure a persistent storage job before %s."
                ),
                job.predecessor.name,
                job.name,
            )
            raise RuntimeError("Invalid storage configuration")
        await utils.symlink(job.file.path, new_path)
        # Symlinks should not be used as a basis for further processing.
        # We therefore only emit the symlinked file but continue with the original.
        if config.emit:
            job.emit_file(stored_file)
        job.next()
        return

    if job.file.keep:
        await utils.copy_file(job.file.path, new_path)
    else:
        await utils.rename_file(job.file.path, new_path)
    job.next(stored_file, emit=config.emit)


class StoreByHash(HashFile, StoreFile):
    alg: Literal[*_hashes] | None = Field(
        default=None,
        description=(
            "Hash algorithm to use. "
            "If unset the job will try to re-use an existing file hash calculated "
            "in a `hash_file` job. "
            "Otherwise, a new hash will be calculated and stored along with the file. "
            "In case the file already has a hash but uses a different algorithm, "
            "a new hash will be calculated for storage but the hash on the file will be kept."
        ),
    )
    levels: int = Field(
        default=2,
        description="Number of directory levels that should be created for stored files.",
    )
    chars_per_level: int = Field(
        default=2,
        description="Number of hash-characters per level.",
    )


@pipeline_task
async def store_by_content_hash(job: Job, config: StoreByHash, file: LocalFile):
    """
    Stores the job file in a directory-tree based on the file’s content hash.

    The generated directory structure looks like this:

    ```
    audio_dir/
      8d/
        a9/
          8da9bce68a6aebdcba325cf21402c78c1628c9da1278b817a600cdd92b720653.flac
          8da9fc4939da378a720ba1ba310d3d7a1a85e44b79cd4c68bfc4bd3081f01062.flac
      e0/
        0a/
          e00afc4939da378a720ba1ba310d3d7a1a85e44b79cd4c68bfc4bd3081f01062.flac
    ```
    """

    file_hash: str | None = None
    store_hash: str | None = None
    if file.hash is not None:
        file_hash = file.hash
        alg, store_hash = file_hash.split(":", 1)
        if config.alg is not None and config.alg != alg:
            store_hash = None
    if store_hash is None:
        _file_hash = await utils.hash_file(file.path, alg=config.alg)
        _, store_hash = _file_hash.split(":", 1)
        if file_hash is None:
            file_hash = _file_hash

    storage_dir = config.storage_dir
    for level in range(config.levels):
        start = level * config.chars_per_level
        end = start + config.chars_per_level
        storage_dir /= store_hash[start:end]
    ext = file.path.suffix
    await store_file(job, config, storage_dir / f"{store_hash}{ext}", hash=file_hash)


class StorageByUploadDate(StoreFile):
    month: bool = Field(
        default=True,
        description="Include month in storage subdirectories.",
    )
    day: bool = Field(
        default=True,
        description="Include day in storage subdirectories.",
    )


@pipeline_task
async def store_by_creation_date(job: Job, config: StorageByUploadDate, file: LocalFile):
    """
    Stores the job file in a directory-tree based on the creation date.

    The generated directory structure looks like this:

    ```
    audio_dir/
      2025/
        11/
          20/
            filename2.flac
            filename3.flac
        10/
          16/
            filename1.flac
    ```
    """
    upload_dt = job.file.created_at
    storage_dir = config.storage_dir
    storage_dir /= str(upload_dt.year)
    if config.month:
        storage_dir /= str(upload_dt.month)
    if config.day:
        storage_dir /= str(upload_dt.day)
    await store_file(job, config, storage_dir / file.path.name, hash=file.hash)


@pipeline_task
async def discard_remote(job: Job):
    """
    Stops the pipeline if the current file is not a local file.
    """
    exists = isinstance(job.file, LocalFile) and await aiofiles.os.path.exists(job.file.path)
    if not exists:
        logger.info("Skipping %s. Not a local file.", job.file.path)
        return
    job.next()


class KeepFile(BaseModel):
    def keeps_file(self):
        return True


@pipeline_task
async def keep_file(job: Job, config: KeepFile):
    """
    Keep the current job file and do not delete it after
    the pipeline finishes.

    This can be helpful to debug intermediate job output.
    """
    job.next(job.file.clone(keep=True))


class LogFileSize(BaseModel):
    force_update_reference: bool = Field(
        default=False,
        description=(
            "If set to true, the reference used to calculate changes in the file size "
            "between jobs will be updated. Implicitly `true` on first call."
        ),
    )


@pipeline_task
async def log_file_size(job: Job, config: LogFileSize, file: LocalFile):
    """
    Logs the file size in MiB.

    If used multiple times in the same pipeline config, it will also
    display the change in file size compared to the reference.
    """
    stat = await aiofiles.os.stat(file.path)
    size = stat.st_size / 1024 / 1024
    orig_size = job.context.get(log_file_size, None)
    context = {log_file_size: size}
    if orig_size is None:
        msg = f"File size: {size:.2f}MiB"
        logger.info(msg)
        job.attach_logs(msg)
    else:
        reduction = ((size - orig_size) / orig_size) * 100
        msg = f"File size: {size:.2f}MiB ({reduction:.2f}% compared to reference)"
        logger.info(msg)
        job.attach_logs(msg)
        if not config.force_update_reference:
            context = None

    job.next(context=context)



from collections.abc import Collection
import fnmatch
import logging
import mimetypes
from pathlib import Path, PurePath
from typing import Literal

import aiofiles
import httpx

from filerohr import config, utils
from filerohr.pipeline.types import OnProgress

logger = logging.getLogger(__name__)


class DisallowedSchemeError(ValueError):
    pass


async def _is_local_file(path: Path) -> bool:
    return await aiofiles.os.path.exists(path)


def _mib_to_bytes(size_in_mib: float) -> int:
    return int(size_in_mib * 1024 * 1024)


def _bytes_to_mib(size_in_bytes: int) -> float:
    return size_in_bytes / 1024 / 1024


def _validate_content_type(response: httpx.Response, content_type_filter: set[str] | None):
    if content_type_filter is None:
        return None
    try:
        content_type = response.headers.get("Content-Type", None).split(";")[0].strip()
    except AttributeError:
        content_type = None
    if content_type is None and "pass_unset" not in content_type_filter:
        raise ValueError("No Content-Type presented.")
    if not any(fnmatch.fnmatch(content_type, pattern) for pattern in content_type_filter):
        raise ValueError(f"Content-Type {content_type} not allowed.")
    return content_type


def _validate_response_size(response: httpx.Response, max_size_mib: float, allow_streams: bool):
    try:
        content_length_bytes = int(response.headers.get("Content-Length", None))
    except TypeError:
        content_length_bytes = None
    if content_length_bytes is None and not allow_streams:
        raise ValueError("Streaming downloads are not allowed.")
    max_size_bytes = _mib_to_bytes(max_size_mib)
    if max_size_mib != 0 and content_length_bytes > max_size_bytes:
        raise ValueError(
            f"Download size {_bytes_to_mib(content_length_bytes):.2f}MiB is larger "
            f"than allowed size {max_size_mib:.2f}MiB."
        )
    return content_length_bytes


async def download_file(
    url: str,
    *,
    follow_redirects: bool = True,
    timeout_seconds: float | int | None = None,
    match_content_type: Collection[str | Literal["pass_unset"]] | None = None,
    max_download_size_mib: float = 0,
    allow_streams: bool = False,
    allowed_schemes: Collection[Literal["http", "https"]] | None = None,
    storage_dir: Path | None = None,
    download_chunk_size_mib: float = 1,
    on_progress: OnProgress | None = None,
):
    if storage_dir is None:
        storage_dir = config.DATA_DIR / "downloads"

    scheme = httpx.URL(url).scheme
    allowed_schemes = set(allowed_schemes if allowed_schemes is not None else {"http", "https"})
    if scheme not in allowed_schemes:
        raise DisallowedSchemeError(scheme)

    client = httpx.AsyncClient()
    logger.info("Downloading %s.", url)
    try:
        response = await client.get(
            url,
            follow_redirects=follow_redirects,
            timeout=timeout_seconds,
        )
        response.raise_for_status()
    except httpx.TimeoutException as exc:
        raise TimeoutError("Download timed out.") from exc

    # pre-download-checks
    if match_content_type is not None:
        match_content_type = set(match_content_type)
    _validate_content_type(response, match_content_type)
    content_length = _validate_response_size(response, max_download_size_mib, allow_streams)

    # start download
    downloaded_bytes = 0
    max_size_bytes = _mib_to_bytes(max_download_size_mib)
    async with aiofiles.tempfile.NamedTemporaryFile(
        "wb", delete_on_close=False, delete=False, dir=storage_dir
    ) as tmp_file:
        for chunk in response.iter_bytes(_mib_to_bytes(download_chunk_size_mib)):
            downloaded_bytes += len(chunk)
            downloaded_mib = _bytes_to_mib(downloaded_bytes)
            if downloaded_bytes > max_size_bytes:
                raise ValueError(
                    f"Download size {downloaded_mib:.2f}MiB is larger than "
                    f"allowed size {max_download_size_mib:.2f}MiB."
                )
            await tmp_file.write(chunk)
            if content_length and on_progress:
                on_progress(downloaded_bytes / content_length * 100)
        logger.info("Downloaded %s with a size of %.2fMiB.", url, downloaded_mib)

    # add file extension
    path = Path(tmp_file.name)
    mime_type = await utils.get_mime_type(path)
    ext = None
    url_path = PurePath(url)
    if mime_type:
        ext = mimetypes.guess_extension(mime_type)
    if not ext:
        ext = url_path.suffix
    if ext:
        new_path = path.with_suffix(ext)
        await utils.rename_file(path, new_path)
        path = new_path

    return path, url_path.name, mime_type



def cast_bool(value: str | None | bool | int) -> bool:
    if value is None:
        return False
    if isinstance(value, bool):
        return value
    return str(value).lower() in ("true", "t", "1", "on", "enabled", "yes", "y")



from .generic import *  # noqa: F403
from .http import *  # noqa: F403
from .io import *  # noqa: F403
from .time import *  # noqa: F403
from .validator import *  # noqa: F403
from .yaml import *  # noqa: F403



from contextlib import asynccontextmanager
import hashlib
import mimetypes
import os
from pathlib import Path
import shutil

import aiofiles
import aiofiles.os
from pebble import asynchronous


@asynchronous.thread
def mkfifo(path: Path):
    os.mkfifo(path)


@asynccontextmanager
async def create_tmp_file(*, delete_on_raise: bool = True, **kwargs):
    async with aiofiles.tempfile.NamedTemporaryFile(**kwargs) as tmp_file:
        try:
            yield Path(tmp_file.name)
        except Exception:
            if delete_on_raise:
                await remove_file(tmp_file.name)
            raise


@asynccontextmanager
async def create_fifo(path: Path):
    try:
        await mkfifo(path)
        yield path
    finally:
        await remove_file(path)


async def remove_file(path: Path):
    try:
        await aiofiles.os.unlink(path)
    except FileNotFoundError:
        pass


@asynchronous.thread
def rename_file(src: Path, dst: Path):
    dst.parent.mkdir(parents=True, exist_ok=True)
    shutil.move(src, dst)


@asynchronous.thread
def copy_file(src: Path, dst: Path):
    dst.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy2(src, dst)


@asynchronous.thread
def get_mime_type(path: Path):
    mime, _ = mimetypes.guess_file_type(path)
    return mime


async def symlink(pointing_to: Path, symlink_path: Path):
    await aiofiles.os.makedirs(symlink_path.parent, exist_ok=True)
    try:
        await aiofiles.os.symlink(pointing_to, symlink_path)
    except FileExistsError as exc:
        if Path(exc.filename) == pointing_to:
            return
        raise


@asynchronous.process
def hash_file(path: Path, *, alg: str = "sha256", prefix: bool = True):
    with path.open("rb") as f:
        file_hash = hashlib.file_digest(f, alg).hexdigest()
        return f"{alg}:{file_hash}" if prefix else file_hash
